package queue

import (
	"git.gammaspectra.live/S.O.N.G/Kirika/audio"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/filter"
	"log"
	"sync"
	"sync/atomic"
)

type QueueIdentifier int

type QueueEntry struct {
	Identifier     QueueIdentifier
	Source         audio.Source
	ReadSamples    int
	cancel         chan bool
	StartCallback  func(q *Queue, entry *QueueEntry)
	EndCallback    func(q *Queue, entry *QueueEntry)
	RemoveCallback func(q *Queue, entry *QueueEntry)
}

type Queue struct {
	queue             []*QueueEntry
	float32Output     *audio.Float32Source
	int16Output       *audio.Int16Source
	int32Output       *audio.Int32Source
	interrupt         chan bool
	interruptDepth    int64
	closed            bool
	lock              sync.RWMutex
	wg                sync.WaitGroup
	identifierCounter QueueIdentifier
}

func NewQueue(format audio.SourceFormat, bitDepth, sampleRate, channels int) *Queue {
	if channels != 1 && channels != 2 {
		log.Panicf("not allowed channel number %d", channels)
	}

	q := &Queue{
		interrupt: make(chan bool, 1),
	}

	switch format {
	case audio.SourceFloat32:
		q.float32Output = audio.NewFloat32Source(sampleRate, channels)
	case audio.SourceInt16:
		q.int16Output = audio.NewInt16Source(bitDepth, sampleRate, channels)
	case audio.SourceInt32:
		q.int32Output = audio.NewInt32Source(bitDepth, sampleRate, channels)
	default:
		log.Panicf("not found source format %d", int(format))
	}

	q.start()
	return q
}

func (q *Queue) spliceSources(input audio.Source) (output audio.Source, cancel chan bool) {
	cancel = make(chan bool, 1)

	switch q.GetSource().GetFormat() {
	case audio.SourceFloat32:
		output = audio.NewFloat32Source(input.GetSampleRate(), input.GetChannels())

		sourceChannel := input.ToFloat32().GetBlocks()
		go func() {
			defer output.Close()
		L1:
			for {
				select {
				case <-cancel:
					break L1
				case block, more := <-sourceChannel:
					if !more {
						//no more blocks!
						break L1
					} else {
						output.IngestFloat32(block)
					}
				}
			}

			input.Unlock()

			//sink remaining
			go audio.NewNullSink().Process(input)
		}()
	case audio.SourceInt16:
		bitDepth := input.GetBitDepth()
		if bitDepth > 16 {
			bitDepth = 16
		}
		output = audio.NewInt16Source(bitDepth, input.GetSampleRate(), input.GetChannels())

		sourceChannel := input.ToInt16().GetBlocks()
		go func() {
			defer output.Close()
		L2:
			for {
				select {
				case <-cancel:
					break L2
				case block, more := <-sourceChannel:
					if !more {
						//no more blocks!
						break L2
					} else {
						output.IngestInt16(block, bitDepth)
					}
				}
			}

			input.Unlock()

			//sink remaining
			go audio.NewNullSink().Process(input)
		}()
	case audio.SourceInt32:
		output = audio.NewInt32Source(input.GetBitDepth(), input.GetSampleRate(), input.GetChannels())

		sourceChannel := input.ToInt32(input.GetBitDepth()).GetBlocks()
		go func() {
			defer output.Close()
		L3:
			for {
				select {
				case <-cancel:
					break L3
				case block, more := <-sourceChannel:
					if !more {
						//no more blocks!
						break L3
					} else {
						output.IngestInt32(block, output.GetBitDepth())
					}
				}
			}

			input.Unlock()

			//sink remaining
			go audio.NewNullSink().Process(input)
		}()
	default:
		log.Panicf("not found source format %d", int(input.GetFormat()))
	}

	return
}

func (q *Queue) start() {
	q.wg.Add(1)
	go func() {
		defer q.wg.Done()
		var current *QueueEntry

		if q.float32Output != nil {
		L1:
			for {
				q.lock.RLock()
				if q.closed {
					q.float32Output.Close()
					break L1
				}
				if len(q.queue) == 0 { //no more entries, wait for interrupt
					q.lock.RUnlock()
					<-q.interrupt
					atomic.AddInt64(&q.interruptDepth, -1)
					continue
				}
				current = q.queue[0]
				q.lock.RUnlock()

			F1:
				for {
					select {
					case <-q.interrupt:
						atomic.AddInt64(&q.interruptDepth, -1)
						//force recheck
						break F1
					case block, more := <-current.Source.(*audio.Float32Source).Blocks:
						if !more {
							//no more blocks! skip
							if current.EndCallback != nil {
								current.EndCallback(q, current)
							}
							q.Remove(current.Identifier)
							break F1
						} else {
							if current.StartCallback != nil && current.ReadSamples == 0 && len(block) > 0 {
								current.StartCallback(q, current)
							}
							current.ReadSamples += len(block) / current.Source.GetChannels()
							q.float32Output.IngestFloat32(block)
						}
					}
				}

			}
		} else if q.int16Output != nil {
		L2:
			for {
				q.lock.RLock()
				if q.closed {
					q.int16Output.Close()
					break L2
				}
				if len(q.queue) == 0 { //no more entries, wait for interrupt
					q.lock.RUnlock()
					<-q.interrupt
					atomic.AddInt64(&q.interruptDepth, -1)
					continue
				}
				current = q.queue[0]
				q.lock.RUnlock()

			F2:
				for {
					select {
					case <-q.interrupt:
						atomic.AddInt64(&q.interruptDepth, -1)
						//force recheck
						break F2
					case block, more := <-current.Source.(*audio.Int16Source).Blocks:
						if !more {
							//no more blocks! skip
							if current.EndCallback != nil {
								current.EndCallback(q, current)
							}
							q.Remove(current.Identifier)
							break F2
						} else {
							if current.StartCallback != nil && current.ReadSamples == 0 && len(block) > 0 {
								current.StartCallback(q, current)
							}
							current.ReadSamples += len(block) / current.Source.GetChannels()
							q.int16Output.IngestInt16(block, current.Source.GetBitDepth())
						}
					}
				}

			}
		} else if q.int32Output != nil {
		L3:
			for {
				q.lock.RLock()
				if q.closed {
					q.int32Output.Close()
					break L3
				}
				if len(q.queue) == 0 { //no more entries, wait for interrupt
					q.lock.RUnlock()
					<-q.interrupt
					atomic.AddInt64(&q.interruptDepth, -1)
					continue
				}
				current = q.queue[0]
				q.lock.RUnlock()

			F3:
				for {
					select {
					case <-q.interrupt:
						atomic.AddInt64(&q.interruptDepth, -1)
						//force recheck
						break F3
					case block, more := <-current.Source.(*audio.Int32Source).Blocks:
						if !more {
							//no more blocks! skip
							if current.EndCallback != nil {
								current.EndCallback(q, current)
							}
							q.Remove(current.Identifier)
							break F3
						} else {
							if current.StartCallback != nil && current.ReadSamples == 0 && len(block) > 0 {
								current.StartCallback(q, current)
							}
							current.ReadSamples += len(block) / current.Source.GetChannels()
							q.int32Output.IngestInt32(block, current.Source.GetBitDepth())
						}
					}
				}

			}
		}

	}()

}
func (q *Queue) getFilterChain(source audio.Source) audio.Source {
	if q.GetChannels() == 1 {
		return filter.NewFilterChain(source, filter.MonoFilter{}, filter.NewResampleFilter(q.GetSampleRate(), filter.QualityFastest, 0), filter.SourceFormatFilter{
			Format: q.GetSource().GetFormat(),
		})
	} else {
		return filter.NewFilterChain(source, filter.StereoFilter{}, filter.NewResampleFilter(q.GetSampleRate(), filter.QualityFastest, 0), filter.SourceFormatFilter{
			Format: q.GetSource().GetFormat(),
		})
	}
}

func (q *Queue) AddHead(source audio.Source, startCallback func(q *Queue, entry *QueueEntry), endCallback func(q *Queue, entry *QueueEntry), removeCallback func(q *Queue, entry *QueueEntry)) (identifier QueueIdentifier) {
	q.lock.Lock()

	splicedOutput, cancel := q.spliceSources(source)
	identifier = q.identifierCounter
	if len(q.queue) > 0 {
		q.queue = append(q.queue[:1], append([]*QueueEntry{{
			Identifier:     identifier,
			Source:         q.getFilterChain(splicedOutput),
			cancel:         cancel,
			StartCallback:  startCallback,
			EndCallback:    endCallback,
			RemoveCallback: removeCallback,
		}}, q.queue[1:]...)...)
	} else {
		q.queue = append(q.queue, &QueueEntry{
			Identifier:     identifier,
			Source:         q.getFilterChain(splicedOutput),
			cancel:         cancel,
			StartCallback:  startCallback,
			EndCallback:    endCallback,
			RemoveCallback: removeCallback,
		})
	}
	q.identifierCounter++
	q.lock.Unlock()
	q.sendInterrupt()

	return
}

func (q *Queue) AddTail(source audio.Source, startCallback func(q *Queue, entry *QueueEntry), endCallback func(q *Queue, entry *QueueEntry), removeCallback func(q *Queue, entry *QueueEntry)) (identifier QueueIdentifier) {
	q.lock.Lock()

	splicedOutput, cancel := q.spliceSources(source)
	identifier = q.identifierCounter
	q.queue = append(q.queue, &QueueEntry{
		Identifier:     identifier,
		Source:         q.getFilterChain(splicedOutput),
		cancel:         cancel,
		StartCallback:  startCallback,
		EndCallback:    endCallback,
		RemoveCallback: removeCallback,
	})
	q.identifierCounter++
	q.lock.Unlock()
	q.sendInterrupt()

	return
}

func (q *Queue) IsClosed() bool {
	q.lock.RLock()
	defer q.lock.RUnlock()
	return q.closed
}

func (q *Queue) Remove(identifier QueueIdentifier) bool {
	q.lock.Lock()

	for i, e := range q.queue {
		if e.Identifier == identifier {
			q.sendInterrupt()
			e.cancel <- true

			e.Source.Unlock()

			go audio.NewNullSink().Process(e.Source)
			//delete entry
			q.queue = append(q.queue[:i], q.queue[i+1:]...)
			q.lock.Unlock()
			if e.RemoveCallback != nil {
				e.RemoveCallback(q, e)
			}
			return true
		}
	}
	q.lock.Unlock()

	return false

}

func (q *Queue) GetQueueHead() *QueueEntry {
	q.lock.RLock()
	defer q.lock.RUnlock()
	if len(q.queue) > 0 {
		return q.queue[0]
	}
	return nil
}

func (q *Queue) GetQueueTail() (index int, entry *QueueEntry) {
	q.lock.RLock()
	defer q.lock.RUnlock()
	if len(q.queue) > 0 {
		return len(q.queue) - 1, q.queue[len(q.queue)-1]
	}
	return 0, nil
}

func (q *Queue) GetQueueIndex(index int) *QueueEntry {
	q.lock.RLock()
	defer q.lock.RUnlock()
	if len(q.queue) > index {
		return q.queue[index]
	}
	return nil
}

func (q *Queue) GetQueueEntry(identifier QueueIdentifier) (index int, entry *QueueEntry) {
	q.lock.RLock()
	defer q.lock.RUnlock()
	for i, e := range q.queue {
		if e.Identifier == identifier {
			return i, e
		}
	}
	return -1, nil
}

func (q *Queue) GetQueueSize() int {
	q.lock.RLock()
	defer q.lock.RUnlock()

	return len(q.queue)
}

func (q *Queue) GetQueue() (entries []*QueueEntry) {
	q.lock.RLock()
	defer q.lock.RUnlock()

	entries = make([]*QueueEntry, len(q.queue))
	copy(entries, q.queue)

	return
}

func (q *Queue) GetSource() audio.Source {
	if q.float32Output != nil {
		return q.float32Output
	} else if q.int16Output != nil {
		return q.int16Output
	} else if q.int32Output != nil {
		return q.int32Output
	}
	return nil
}

func (q *Queue) GetSampleRate() int {
	return q.GetSource().GetSampleRate()
}

func (q *Queue) GetChannels() int {
	return q.GetSource().GetChannels()
}

func (q *Queue) sendInterrupt() {
	//TODO: maybe use len() on channel?
	if atomic.LoadInt64(&q.interruptDepth) == 0 { //not waiting on interrupt
		atomic.AddInt64(&q.interruptDepth, 1)
		q.interrupt <- true
	}
}

func (q *Queue) Close() {
	if !q.closed {
		q.closed = true
		q.sendInterrupt()
	}
}

func (q *Queue) Wait() {
	q.wg.Wait()
}
