package audio

import (
	"git.gammaspectra.live/S.O.N.G/Kirika/vector"
)

type Float32Source struct {
	SampleRate int
	Channels   int
	Blocks     chan []float32
	locked     bool
}

func NewFloat32Source(sampleRate, channels int) *Float32Source {
	return &Float32Source{
		SampleRate: sampleRate,
		Channels:   channels,
		Blocks:     make(chan []float32),
	}
}

func (s *Float32Source) Split(n int) (sources []Source) {
	if s.Locked() {
		return
	}
	sources = make([]Source, n)
	for i := range sources {
		copiedSource := *s
		copiedSource.Blocks = make(chan []float32, cap(s.Blocks))
		sources[i] = &copiedSource
	}

	go func() {
		defer func() {
			for _, source := range sources {
				source.Close()
			}
		}()

		for block := range s.GetBlocks() {
			for _, source := range sources {
				source.IngestFloat32(block)
			}
		}
	}()

	return
}

func (s *Float32Source) GetBlocks() chan []float32 {
	if s.Locked() {
		return nil
	}
	s.locked = true
	return s.Blocks
}

func (s *Float32Source) Close() {
	close(s.Blocks)
}

func (s *Float32Source) GetSampleRate() int {
	return s.SampleRate
}

func (s *Float32Source) GetChannels() int {
	return s.Channels
}

func (s *Float32Source) GetBitDepth() int {
	return 0
}

func (s *Float32Source) GetFormat() SourceFormat {
	return SourceFloat32
}

func (s *Float32Source) Locked() bool {
	return s.locked
}

func (s *Float32Source) Unlock() {
	s.locked = false
}

func (s *Float32Source) ToFloat32() *Float32Source {
	return s
}

func (s *Float32Source) ToInt16() *Int16Source {
	if s.Locked() {
		return nil
	}
	source := &Int16Source{
		SampleRate: s.SampleRate,
		Channels:   s.Channels,
		BitDepth:   16,
		Blocks:     make(chan []int16),
	}

	go func() {
		defer source.Close()
		for block := range s.GetBlocks() {
			source.Blocks <- vector.Float32ToInt16(block)
		}
	}()

	return source
}

func (s *Float32Source) ToInt32(bitDepth int) *Int32Source {
	if s.Locked() {
		return nil
	}
	source := &Int32Source{
		SampleRate: s.SampleRate,
		Channels:   s.Channels,
		BitDepth:   bitDepth,
		Blocks:     make(chan []int32),
	}

	go func() {
		defer source.Close()
		for block := range s.GetBlocks() {
			source.Blocks <- vector.Float32ToInt32(block, bitDepth)
		}
	}()

	return source
}

func (s *Float32Source) IngestFloat32(buf []float32) {
	s.Blocks <- buf
}

func (s *Float32Source) IngestInt8(buf []int8, bitDepth int) {
	s.Blocks <- vector.Int8ToFloat32(buf, bitDepth)
}

func (s *Float32Source) IngestInt16(buf []int16, bitDepth int) {
	s.Blocks <- vector.Int16ToFloat32(buf, bitDepth)
}

func (s *Float32Source) IngestInt24(buf []byte, bitDepth int) {
	s.Blocks <- vector.Int24ToFloat32(buf, bitDepth)
}

func (s *Float32Source) IngestInt32(buf []int32, bitDepth int) {
	s.Blocks <- vector.Int32ToFloat32(buf, bitDepth)
}

func (s *Float32Source) GetBlocksInterface() interface{} {
	return s.GetBlocks()
}

type Int16Source struct {
	BitDepth   int
	SampleRate int
	Channels   int
	Blocks     chan []int16
	locked     bool
}

func NewInt16Source(bitDepth, sampleRate, channels int) *Int16Source {
	if bitDepth <= 0 || bitDepth > 16 {
		return nil
	}
	return &Int16Source{
		BitDepth:   bitDepth,
		SampleRate: sampleRate,
		Channels:   channels,
		Blocks:     make(chan []int16),
	}
}

func (s *Int16Source) Split(n int) (sources []Source) {
	if s.Locked() {
		return
	}
	sources = make([]Source, n)
	for i := range sources {
		copiedSource := *s
		copiedSource.Blocks = make(chan []int16, cap(s.Blocks))
		sources[i] = &copiedSource
	}

	go func() {
		defer func() {
			for _, source := range sources {
				source.Close()
			}
		}()

		for block := range s.GetBlocks() {
			for _, source := range sources {
				source.IngestInt16(block, s.BitDepth)
			}
		}
	}()

	return
}

func (s *Int16Source) GetBlocks() chan []int16 {
	if s.Locked() {
		return nil
	}
	s.locked = true
	return s.Blocks
}

func (s *Int16Source) Close() {
	close(s.Blocks)
}

func (s *Int16Source) GetSampleRate() int {
	return s.SampleRate
}

func (s *Int16Source) GetChannels() int {
	return s.Channels
}

func (s *Int16Source) GetBitDepth() int {
	return s.BitDepth
}

func (s *Int16Source) GetFormat() SourceFormat {
	return SourceInt16
}

func (s *Int16Source) Locked() bool {
	return s.locked
}

func (s *Int16Source) Unlock() {
	s.locked = false
}

func (s *Int16Source) ToFloat32() *Float32Source {
	if s.Locked() {
		return nil
	}

	source := &Float32Source{
		SampleRate: s.SampleRate,
		Channels:   s.Channels,
		Blocks:     make(chan []float32),
	}

	go func() {
		defer source.Close()
		for block := range s.GetBlocks() {
			source.Blocks <- vector.Int16ToFloat32(block, s.BitDepth)
		}
	}()

	return source
}

func (s *Int16Source) ToInt16() *Int16Source {
	return s
}

func (s *Int16Source) ToInt32(bitDepth int) *Int32Source {
	if s.Locked() {
		return nil
	}

	source := &Int32Source{
		SampleRate: s.SampleRate,
		Channels:   s.Channels,
		BitDepth:   bitDepth,
		Blocks:     make(chan []int32),
	}

	go func() {
		diff := s.BitDepth - bitDepth

		if diff >= 0 {
			defer source.Close()
			for block := range s.GetBlocks() {
				buf := make([]int32, len(block))
				for i := range block {
					buf[i] = int32(block[i]) >> diff
				}
				source.Blocks <- buf
			}
		} else {
			diff = -diff
			defer source.Close()
			for block := range s.GetBlocks() {
				buf := make([]int32, len(block))
				for i := range block {
					buf[i] = int32(block[i]) << diff
				}
				source.Blocks <- buf
			}
		}
	}()

	return source
}

func (s *Int16Source) IngestFloat32(buf []float32) {
	s.Blocks <- vector.Float32ToInt16(buf)
}

func (s *Int16Source) IngestInt8(buf []int8, bitDepth int) {
	block := make([]int16, len(buf))
	for i := range buf {
		block[i] = int16(buf[i]) << (s.BitDepth - bitDepth)
	}
	s.Blocks <- block
}

func (s *Int16Source) IngestInt16(buf []int16, bitDepth int) {
	if bitDepth == s.BitDepth {
		s.Blocks <- buf
	} else {
		block := make([]int16, len(buf))
		for i := range buf {
			block[i] = buf[i] << (s.BitDepth - bitDepth)
		}
		s.Blocks <- block
	}
}

func (s *Int16Source) IngestInt24(buf []byte, bitDepth int) {
	block := make([]int16, len(buf)/3)

	for i := 0; i < len(buf); i += 3 {
		sample := uint32(buf[i])
		sample += uint32(buf[i+1]) << 8
		sample += uint32(buf[i+2]) << 16

		//sign extend
		block[i/3] = int16(int32(sample<<8) >> (bitDepth - s.BitDepth + 8))
	}
	s.Blocks <- block
}

func (s *Int16Source) IngestInt32(buf []int32, bitDepth int) {
	block := make([]int16, len(buf))
	for i := range buf {
		block[i] = int16(buf[i] >> (bitDepth - s.BitDepth))
	}
	s.Blocks <- block
}

func (s *Int16Source) GetBlocksInterface() interface{} {
	return s.GetBlocks()
}

type Int32Source struct {
	BitDepth   int
	SampleRate int
	Channels   int
	Blocks     chan []int32
	locked     bool
}

func NewInt32Source(bitDepth, sampleRate, channels int) *Int32Source {
	if bitDepth <= 0 || bitDepth > 32 {
		return nil
	}
	return &Int32Source{
		BitDepth:   bitDepth,
		SampleRate: sampleRate,
		Channels:   channels,
		Blocks:     make(chan []int32),
	}
}

func (s *Int32Source) Split(n int) (sources []Source) {
	if s.Locked() {
		return
	}
	sources = make([]Source, n)
	for i := range sources {
		copiedSource := *s
		copiedSource.Blocks = make(chan []int32, cap(s.Blocks))
		sources[i] = &copiedSource
	}

	go func() {
		defer func() {
			for _, source := range sources {
				source.Close()
			}
		}()

		for block := range s.GetBlocks() {
			for _, source := range sources {
				source.IngestInt32(block, s.BitDepth)
			}
		}
	}()

	return
}

func (s *Int32Source) GetBlocks() chan []int32 {
	if s.Locked() {
		return nil
	}
	s.locked = true
	return s.Blocks
}

func (s *Int32Source) Unlock() {
	s.locked = false
}

func (s *Int32Source) Close() {
	close(s.Blocks)
}

func (s *Int32Source) GetSampleRate() int {
	return s.SampleRate
}

func (s *Int32Source) GetChannels() int {
	return s.Channels
}

func (s *Int32Source) GetBitDepth() int {
	return s.BitDepth
}

func (s *Int32Source) GetFormat() SourceFormat {
	return SourceInt32
}

func (s *Int32Source) Locked() bool {
	return s.locked
}

func (s *Int32Source) ToInt16() *Int16Source {
	if s.Locked() {
		return nil
	}

	source := &Int16Source{
		SampleRate: s.SampleRate,
		Channels:   s.Channels,
		BitDepth:   16,
		Blocks:     make(chan []int16),
	}

	go func() {
		defer source.Close()

		for block := range s.GetBlocks() {
			source.IngestInt16(vector.Int32ToInt16(block, s.BitDepth), 16)
		}
	}()

	return source
}

func (s *Int32Source) ToInt32(bitDepth int) *Int32Source {
	if s.BitDepth == bitDepth {
		return s
	}

	source := &Int32Source{
		SampleRate: s.SampleRate,
		Channels:   s.Channels,
		BitDepth:   bitDepth,
		Blocks:     make(chan []int32),
	}

	go func() {
		diff := s.BitDepth - bitDepth
		defer source.Close()

		if diff >= 0 {
			for block := range s.GetBlocks() {
				buf := make([]int32, len(block))
				for i := range block {
					buf[i] = block[i] >> diff
				}
				source.IngestInt32(buf, bitDepth)
			}
		} else {
			diff = -diff
			for block := range s.GetBlocks() {
				buf := make([]int32, len(block))
				for i := range block {
					buf[i] = block[i] << diff
				}
				source.IngestInt32(buf, bitDepth)
			}
		}
	}()

	return source
}

func (s *Int32Source) ToFloat32() *Float32Source {
	if s.Locked() {
		return nil
	}

	source := &Float32Source{
		SampleRate: s.SampleRate,
		Channels:   s.Channels,
		Blocks:     make(chan []float32),
	}

	go func() {
		defer source.Close()
		for block := range s.GetBlocks() {
			source.Blocks <- vector.Int32ToFloat32(block, s.BitDepth)
		}
	}()

	return source
}

func (s *Int32Source) IngestFloat32(buf []float32) {
	s.Blocks <- vector.Float32ToInt32(buf, s.BitDepth)
}

func (s *Int32Source) IngestInt8(buf []int8, bitDepth int) {
	block := make([]int32, len(buf))
	for i := range buf {
		block[i] = int32(buf[i]) << (s.BitDepth - bitDepth)
	}
	s.Blocks <- block
}

func (s *Int32Source) IngestInt16(buf []int16, bitDepth int) {
	block := make([]int32, len(buf))
	for i := range buf {
		block[i] = int32(buf[i]) << (s.BitDepth - bitDepth)
	}
	s.Blocks <- block
}

func (s *Int32Source) IngestInt24(buf []byte, bitDepth int) {
	s.Blocks <- vector.BytesToInt32(buf, bitDepth)
}

func (s *Int32Source) IngestInt32(buf []int32, bitDepth int) {
	if bitDepth == s.BitDepth {
		s.Blocks <- buf
	} else {
		block := make([]int32, len(buf))
		for i := range buf {
			block[i] = buf[i] << (s.BitDepth - bitDepth)
		}
		s.Blocks <- block
	}
}

func (s *Int32Source) GetBlocksInterface() interface{} {
	return s.GetBlocks()
}

type SourceFormat int

const (
	SourceFloat32 = SourceFormat(iota)
	SourceInt16
	SourceInt24
	SourceInt32
)

type Source interface {
	GetBitDepth() int
	GetSampleRate() int
	GetChannels() int
	GetFormat() SourceFormat
	ToFloat32() *Float32Source
	ToInt16() *Int16Source
	ToInt32(bitDepth int) *Int32Source

	IngestFloat32(buf []float32)
	IngestInt8(buf []int8, bitDepth int)
	IngestInt16(buf []int16, bitDepth int)
	IngestInt24(buf []byte, bitDepth int)
	IngestInt32(buf []int32, bitDepth int)
	GetBlocksInterface() interface{}

	//Split a Source into multiple ones. After calling this function, the source is locked.
	Split(n int) []Source
	Close()
	Unlock()
	Locked() bool
}
